/* history -- command history mechanism                                     */
/* copyright 1985-6  Michael M Rubenstein                                   */

/* version 1.3  30 Apr 87                                                   */

#include <ctype.h>
#define FALSE           0
#define TRUE            1
#define NULL            ((void *) 0)

struct regs {           unsigned    ax, bx, cx, dx, si, di, ds, es; };

/* use bdos functions directly to save memory                               */
#define conin()         bdos(8)
#define conrdy()        bdos(11)
#define conout(c)       bdos(2, c)
#define strout(s)       bdos(9, s)
#define argbuf          ((unsigned char *) 0x81)

/* some interesting characters                                              */
#define CTL             0x1f
#define NUL             0
#define BEL             (CTL & 'G')
#define BS              (CTL & 'H')
#define LF              (CTL & 'J')
#define CR              (CTL & 'M')
#define CTLL            (CTL & 'L')
#define CTLQ            (CTL & 'Q')
#define CTLT            (CTL & 'T')
#define CTLU            (CTL & 'U')
#define CTLW            (CTL & 'W')
#define CTLX            (CTL & 'X')
#define CTLY            (CTL & 'Y')
#define ESC             (CTL & '[')
#define DEL             0x7f

/* extended characters                                                      */
#define F1              256 + 59
#define F2              256 + 60
#define F3              256 + 61
#define F4              256 + 62
#define F5              256 + 63
#define F6              256 + 64
#define F7              256 + 65
#define F8              256 + 66
#define F9              256 + 67
#define F10             256 + 68
#define HOME            256 + 71
#define UP              256 + 72
#define LEFT            256 + 75
#define RIGHT           256 + 77
#define END             256 + 79
#define DOWN            256 + 80
#define INS             256 + 82
#define KDEL            256 + 83
#define CTLLEFT         256 + 115
#define CTLRIGHT        256 + 116

int                     insert = FALSE;         /* insert mode switch */
unsigned char           line[256];

unsigned                lineoff, lineseg;       /* offset and seg of line */
unsigned char           *cur;                   /* current position in line */
int                     len;                    /* max length of line */
unsigned                startpos, endpos, prevpos;   /* screen positions */
int                     vpage,                  /* video page */
                        maxcols,                /* cols on screen */
                        cursor,                 /* standard cursor */
                        icursor;                /* insert cursor */

unsigned char           *buffers[8];
unsigned char           **hist_switch;          /* 0 if command, 4 other */

#define histbuf         hist_switch[0]          /* beginning */
#define endhistbuf      hist_switch[1]          /* end */
#define first           hist_switch[2]          /* logical first */
#define next            hist_switch[3]          /* logical next */

unsigned char           *curhist;               /* current hist line */

unsigned char           *backhist(), *fwdhist(), *nxt(), *prv();
unsigned char           *sethist();

extern void             *sbrk();

/* We don't need the standard Aztec initialization, so the main routine     */
/* is named Croot to replace it, rather than main                           */
Croot()
{
  unsigned char         *p;
  unsigned              l;

  p = argbuf;

  for (hist_switch = buffers; hist_switch < buffers + 8; hist_switch += 4)
  {
    while (isspace(*p))
      ++p;
    /* get history buffer */
    if ((l = atoi(p)) < 256)
      l = 256;
    else
    if (l > 32767)
      l = 32767;
    if ((histbuf = sbrk(l)) == (char *) -1)
    {
      strout("history: insuff mem\n\r$");
      return;
    }
    endhistbuf = histbuf + l - 1;
    first = next = histbuf + 1;
    *first = *histbuf = '\0';

    while (*p != '\0' && !isspace(*p))
      ++p;
  }

  /* set up TSR */
  setup();
}

/* get a line with editing & history                                        */
getline()
{
  int                   c;
  unsigned char         *p;
  static unsigned       oldseg = 0;

  /* the first time we're called it's from COMMAND.COM.  Save the segment */
  /* so we can recognize calls from COMMAND.COM.  Only those calls will use */
  /* the history mechanism.  Editing will apply to all. */
  if (oldseg == 0)
    oldseg = lineseg;

  hist_switch = buffers + ((lineseg == oldseg) ? 0 : 4);

  /* set up for input */
  len = peekb(lineoff, lineseg);
  memset(cur = line, 0, len);
  insert = FALSE;
  init_pos();
  getvpage();
  curhist = next;

  /* main editing loop */
  for (;;)
  {
    c = getcon();
    switch (c)
    {
      case CR:      /* done with line */
                    setcursor(cursor);
                    conout('\r');
                    setpos(endpos);
                    if (line[0] != '\0')
                      puthist();
                    storeline();
                    return;

      case LEFT:    /* back one character */
                    if (cur != line)
                      backup();
                    break;

      case CTLLEFT: /* back one word */
                    backwd();
                    break;

      case CTLRIGHT:
                    /* forward one word */
                    while (isalnum(*cur))
                      forward();
                    while (*cur != 0 && !isalnum(*cur))
                      forward();
                    break;

      case DEL:
      case BS:      /* backspace and delete */
                    if (cur == line)
                      break;
                    backup();           /* NOTE fall through */
     
      case KDEL:    /* delete current char */
                    if (*cur != '\0')
                      delchr();
                    break;

      case CTLW:
      case F9:      /* delete word left */
                    backwd();
                    /* NOTE fall through */
      case CTLT:
      case F10:     /* delete word right */
                    while (isalnum(*cur))
                      delchr();
                    while (*cur != 0 && !isalnum(*cur))
                      delchr();
                    break;

                    /* get pattern if at beginning else forward */
      case F1:      if (cur == line && line[0] == '\0')
                    {
                      getpat();
                      break;
                    }
      case RIGHT:   /* forward one character */
                    if (*cur != 0)
                      forward();
                    break;

      case INS:     /* insert char */
                    setcursor((insert = !insert) ? icursor : cursor);
                    break;

      case CTLU:
      case CTLX:
      case ESC:     /* delete line */
                    delln();
                    curhist = next;
                    break;

      case CTLY:
      case F8:      /* delete to end of line */
                    while (*cur != 0)
                      delchr();
                    break;

      case HOME:    home();
                    break;

      case F3:      getpat();
      case END:     while (*cur != 0)
                      forward();
                    break;

      case UP:      if ((p = backhist(curhist)) != NULL)
                        gethist(p);
                    break;

      case DOWN:    if (p = fwdhist(curhist))
                      gethist(p);
                    break;

      case CTLL:
      case F7:      search();
                    break;

      case NUL:     /* ignore nulls */
                    break;

      case F2:      getpat();
                    c = getcon();
                    if (*cur == '\0')
                      break;
                    for (p = cur + 1; *p != '\0' && *p != c; ++p)
                      ;
                    if (*p != '\0')
                      while (cur < p)
                        forward();
                    break;

      case F4:      getpat();
                    c = getcon();
                    if (*cur == '\0')
                      break;
                    for (p = cur + 1; *p != '\0'&& *p != c; ++p)
                      ;
                    if (*p != '\0')
                      while (*cur != c)
                        delchr();
                    break;

      case F5:      storeline();
                    break;

      case F6:
      case CTLQ:    c = conin();
                    if (c == 0 && conrdy())
                    {
                      conin();
                      break;
                    }
                    /* NOTE fall through */
      default:      if (c > 255)
                      continue;
                    if (cur - line == len - 1)
                      bell();
                    else
                    {
                      if (insert)
                        inschr();
                      *cur = c;
                      forward();
                      showln();
                    }
    }    
  }
}

/* get extended character from console                                      */
getcon()
{
  int                   c;

  return ((c = conin()) == '\0') ? (conin() + 256) : c;
}

/* get pattern from line buffer                                             */
getpat()
{
  int                   i, n;

  if (cur == line && line[0] == '\0')
  {
    for (i = 0, n = peekb(lineoff + 1, lineseg); i < n; ++i)
      line[i] = peekb(lineoff + 2 + i, lineseg);
    line[i] = '\0';
    home();
    showln();
  }
}

/* store the line                                                           */
storeline()
{
  register unsigned char
                        *p;
  register int          i;

  for (i = 0, p = line; *p != '\0'; ++i, ++p)
    pokeb(lineoff + 2 + i, lineseg, *p);
  pokeb(lineoff + 2 + i, lineseg, CR);
  pokeb(lineoff + 1, lineseg, i);
}

/* back one word                                                            */
backwd()
{
  if (cur == line)
    return;
  backup();
  while (cur != line && !isalnum(*cur))
    backup();
  while (cur != line && isalnum(*(cur - 1)))
    backup();
}

/* delete character                                                         */
delchr()
{
  unsigned char         *p;
  unsigned              pos;

  pos = getpos();

  for (p = cur; (*p = *(p + 1)) != '\0'; ++p)
    pctl(*p);

  pctl(' ');
  pctl(' ');
  setpos(pos);
}

/* insert a character in line                                               */
inschr()
{
  unsigned char         *p;

  for (p = cur; *p != '\0'; ++p)
    ;
  if (p > line + len - 2)
    p = line + len - 2;

  *(p + 1) = '\0';

  while (p > cur)
  {
    *p = *(p - 1);
    --p;
  }
  *cur = ' ';
}

/* delete entire line                                                       */
delln()
{
  home();
  while (*cur != '\0')
  {
    if (*cur < ' ' || *cur == DEL)
      conout(' ');
    conout(' ');
    ++cur;
  }
  home();
  init_pos();
  memset(cur = line, 0, len);
}

/* move cursor to start of line */
home()
{
  cur = line;
  setpos(startpos);
}

/* show line from current position                                          */
showln()
{
  unsigned char         *p;

  if (*cur != 0)
  {
    p = cur;
    while (*cur != '\0')
      forward();
    home();
    while (cur < p)
      forward();
  }
}

/* back up one character                                                    */
backup()
{
  unsigned              pos;
  int                   row, col;

  --cur;
  pos = getpos();
  row = (pos >> 8) & 0xff;
  col = pos & 0xff;
  col -= (*cur < ' ' || *cur == DEL) ? 2 : 1;
  if (col < 0)
  {
    --row;
    col = 79;
  }
  setpos((row << 8) + col);
}

/* forward one character                                                    */
forward()
{
  pctl(*(cur++));
}

/* go to previous history line                                              */
unsigned char *backhist(p)
  unsigned char         *p;
{
  if (p == first)
  {
    bell();
    return NULL;
  }
  p = prv(p);
  while (*(p = prv(p)) != '\0')
    ;
  return nxt(p);
}

/* go to next history line                                                  */
unsigned char *fwdhist(p)
  unsigned char         *p;
{
  if (p == next)
  {
    bell();
    return NULL;
  }
  while (*(p = nxt(p)) != '\0')
    ;
  return nxt(p);
}

/* go to previous history character                                         */
unsigned char *prv(p)
  unsigned char         *p;
{
  return (p == histbuf) ? endhistbuf : --p;
}

/* go to next history character                                             */
unsigned char *nxt(p)
  unsigned char         *p;
{
  return (p == endhistbuf) ? histbuf : ++p;
}

/* make history line current and copy to line buffer                        */
gethist(p)
  unsigned char         *p;
{
  int                   i;

  curhist = p;
  delln();
  cur = line;
  for (i = 0; *p != '\0' && i < len - 1; ++i)
  {
    *cur = *p;
    forward();
    p = nxt(p);
  }
  *cur = '\0';
  home();
}

/* save a history line                                                      */
puthist()
{
  int                   c;
  unsigned char         *p, *q;
  unsigned              i;

  q = line;
  do 
  {
    c = *(next) = *(q++);
    next = nxt(next);
    if (next == first)
      first = fwdhist(nxt(first));
  } while (c != '\0');
  *next = '\0';
}

/* search for history line                                                  */
search()
{
  unsigned char         *p, *q, *r;

  for (p = curhist; (p = backhist(p)) != NULL;)
  {
    for (q = p, r = line;
         q != '\0' && r < cur && tolower(*q) == tolower(*r);
         q = nxt(q), ++r)
      ;
    if (r == cur)
    {
      gethist(p);
      while (cur < r)
        forward();
      return;
    }
  }
}

/* put character to console, converting controls to printables              */
pctl(c)
  int                   c;
{
  if (c == DEL)
  {
    pc('^');
    pc('?');
    return;
  }

  if (c < ' ')
  {
    pc('^');
    pc(c + '@');
    return;
  }

  pc(c);
}

/* initialize positions                                                     */
init_pos()
{
  prevpos = endpos = (startpos = getpos()) & 0xff00;
}

/* put character to console                                                 */
pc(c)
  int                   c;
{
  unsigned              pos;

  conout(c);
  pos = getpos();
  if (pos == endpos && prevpos > endpos)
    startpos -= 0x0100;
  else
  if ((pos & 0x00ff) == 0 & pos > endpos)
    endpos = pos;
  prevpos = pos;
}

/* audible alarm                                                            */
bell()
{
  conout(BEL);
}

/* get various screen & cursor characteristics                              */
getvpage()
{
  struct regs           r;

  r.ax = 0x0f00;
  sysint(0x10, &r, &r);
  vpage = r.bx;
  maxcols = r.ax >> 8;

  r.ax = 0x0300;
  sysint(0x10, &r, &r);
  cursor = r.cx;
  icursor = (((cursor & 0xff) - 4) << 8) + (cursor & 0xff);
}

/* get current screen position                                              */
getpos()
{
  struct regs           r;

  r.ax = 0x0300;
  r.bx = vpage;
  sysint(0x10, &r, &r);
  return r.dx;
}

/* set screen position                                                      */
setpos(p)
  unsigned              p;
{
  struct regs           r;

  r.ax = 0x0200;
  r.bx = vpage;
  r.dx = p;
  sysint(0x10, &r, &r);
  prevpos = p;
}

/* set cursor type                                                          */
setcursor(c)
  unsigned              c;
{
  struct regs           r;

  r.ax = 0x0100;
  r.cx = c;
  sysint(0x10, &r, &r);
}
